@@ -89,6 +89,14 @@ test_that("format for tib_vector works", {
       fill = 1:2
     )
   )
+
+  expect_snapshot({
+    tib_lgl_vec("lgl")
+    tib_int_vec("int")
+    tib_dbl_vec("dbl")
+    tib_chr_vec("chr")
+    tib_date_vec("date")
+  })
 })
 
 test_that("format for tib_chr_date works", {
@@ -170,6 +178,7 @@ test_that("format for tib_row works", {
 test_that("format for tib_variant works", {
   expect_snapshot(tib_variant("a"))
   expect_snapshot(tib_variant("a", fill = tibble(a = 1:2)))
+  expect_snapshot(tib_variant("a", elt_transform = as.character))
 })
 
 test_that("format for tib_df works", {
@@ -201,6 +210,30 @@ test_that("format for tib_df works", {
   )
 })
 
+test_that("format for tib_recursive works", {
+  local_options(cli.num_colors = 1)
+  expect_snapshot(
+    tib_recursive(
+      "data",
+      .children = "children",
+      tib_int("id"),
+      tib_chr("name"),
+    ) %>%
+      print()
+  )
+
+  expect_snapshot(
+    tib_recursive(
+      "data",
+      .children = "children",
+      tib_int("id"),
+      tib_chr("name"),
+      .required = FALSE
+    ) %>%
+      print()
+  )
+})
+
 test_that("prints non-canonical names", {
   expect_snapshot({
     tspec_df(b = tib_int("a")) %>% format()
@@ -254,9 +287,34 @@ test_that("prints arguments of spec_*", {
   )
 })
 
+test_that("prints arguments of tspec_recursive", {
+  expect_equal(
+    format(tspec_recursive(tib_int("a"), .children = "children")),
+    'tspec_recursive(\n  .children = "children",\n  tib_int("a"),\n)'
+  )
+})
+
 test_that("format uses trailing comma", {
   expect_equal(
     format(tib_df("x", a = tib_int("a"))),
     'tib_df(\n  "x",\n  tib_int("a"),\n)'
   )
 })
+
+test_that("special ptypes are cprrectly formatted", {
+  check_native <- function(ptype, ctr) {
+    # native
+    expect_equal(format_ptype(ptype), ctr)
+
+    # non-native
+    class(ptype) <- c("a", class(ptype))
+    expect_equal(format_ptype(ptype), deparse(ptype))
+  }
+
+  check_native(logical(), "logical(0)")
+  check_native(character(), "character(0)")
+  check_native(integer(), "integer(0)")
+  check_native(numeric(), "numeric(0)")
+  check_native(vctrs::new_date(), "vctrs::new_date()")
+  check_native(vctrs::new_duration(), "vctrs::new_duration()")
+})
