@@ -48,7 +48,7 @@ test_that("names are checked", {
   })
 })
 
-test_that("scalar column works", {
+test_that("tib_scalar works", {
   dtt <- vctrs::new_datetime(1)
 
   # can parse
@@ -57,8 +57,8 @@ test_that("scalar column works", {
 
   # errors if required but absent
   expect_snapshot({
-    (expect_error(tib(list(), tib_lgl("x"))))
-    (expect_error(tib(list(), tib_scalar("x", dtt))))
+    (expect_error(tib(list(a = 1), tib_lgl("x"))))
+    (expect_error(tib(list(a = 1), tib_scalar("x", dtt))))
   })
 
   # errors if bad size
@@ -77,6 +77,9 @@ test_that("scalar column works", {
 
   # fallback default works
   expect_equal(tib(list(), tib_lgl("x", required = FALSE)), tibble(x = NA))
+  expect_equal(tib(list(), tib_int("x", required = FALSE)), tibble(x = NA_integer_))
+  expect_equal(tib(list(), tib_dbl("x", required = FALSE)), tibble(x = NA_real_))
+  expect_equal(tib(list(), tib_chr("x", required = FALSE)), tibble(x = NA_character_))
   expect_equal(tib(list(), tib_scalar("x", dtt, required = FALSE)), tibble(x = vctrs::new_datetime(NA_real_)))
 
   # use NA if NULL
@@ -104,7 +107,11 @@ test_that("scalar column works", {
   )
 })
 
-test_that("record objects work", {
+test_that("tib_scalar with record objects work", {
+  skip_on_cran()
+  # skip due to issue:
+  # `attr(actual$x, 'balanced')`:   <NA>
+  # `attr(expected$x, 'balanced')`: TRUE
   x_rcrd <- as.POSIXlt(Sys.time(), tz = "UTC")
 
   expect_equal(
@@ -138,7 +145,7 @@ test_that("record objects work", {
   )
 })
 
-test_that("scalar columns respect ptype_inner", {
+test_that("tib_scalar respect ptype_inner", {
   withr::local_timezone("UTC")
   f <- function(x) {
     stopifnot(is.character(x))
@@ -182,7 +189,7 @@ test_that("scalar columns respect ptype_inner", {
   )
 })
 
-test_that("vector column works", {
+test_that("tib_vector works", {
   dtt <- vctrs::new_datetime(1)
 
   # can parse
@@ -191,8 +198,8 @@ test_that("vector column works", {
 
   # errors if required but absent
   expect_snapshot({
-    (expect_error(tib(list(), tib_lgl_vec("x"))))
-    (expect_error(tib(list(), tib_vector("x", dtt))))
+    (expect_error(tib(list(a = 1), tib_lgl_vec("x"))))
+    (expect_error(tib(list(a = 1), tib_vector("x", dtt))))
   })
 
   # errors if bad type
@@ -209,25 +216,40 @@ test_that("vector column works", {
   # uses NULL for NULL
   expect_equal(tib(list(x = NULL), tib_int_vec("x", fill = 1:2)), tibble(x = list_of(NULL, .ptype = integer())))
 
-  # transform works
+  # elt_transform works
   expect_equal(
-    tib(list(x = c(TRUE, FALSE)), tib_lgl_vec("x", transform = ~ !.x)),
+    tib(list(x = c(TRUE, FALSE)), tib_lgl_vec("x", elt_transform = ~ !.x)),
     tibble(x = list_of(c(FALSE, TRUE)))
   )
   expect_equal(
-    tib(list(x = c(dtt - 1, dtt)), tib_vector("x", dtt, transform = ~ .x + 1)),
+    tib(list(x = c(dtt - 1, dtt)), tib_vector("x", dtt, elt_transform = ~ .x + 1)),
     tibble(x = list_of(c(dtt, dtt + 1)))
   )
+
+  # transform works
+  expect_equal(
+    tib(
+      list(x = c(TRUE, FALSE)),
+      tib_lgl_vec("x", transform = ~ vctrs::new_list_of(purrr::map(.x, `!`), ptype = logical()))
+    ),
+    tibble(x = list_of(c(FALSE, TRUE)))
+  )
+
+  # keeps names
+  expect_equal(
+    tib(list(x = c(a = 1L, b = 2L)), tib_int_vec("x")),
+    tibble(x = list_of(c(a = 1L, b = 2L), .ptype = integer()))
+  )
 })
 
-test_that("vector columns respect ptype_inner", {
+test_that("tib_vector respect ptype_inner", {
   spec <- tspec_df(
     tib_vector(
       "x", Sys.Date(),
       required = FALSE,
       ptype_inner = character(),
       fill = as.Date("2000-01-01"),
-      transform = as.Date
+      elt_transform = as.Date
     ),
   )
 
@@ -259,7 +281,7 @@ test_that("explicit NULL work", {
   )
 })
 
-test_that("vector column respects vector_allows_empty_list", {
+test_that("tib_vector respects vector_allows_empty_list", {
   x <- list(
     list(x = 1),
     list(x = list()),
@@ -276,7 +298,7 @@ test_that("vector column respects vector_allows_empty_list", {
   )
 })
 
-test_that("vector column creates tibble with values_to", {
+test_that("tib_vector creates tibble with values_to", {
   spec <- tib_int_vec("x", values_to = "val")
   expect_equal(
     tib(list(x = 1:2), spec),
@@ -303,7 +325,7 @@ test_that("vector column creates tibble with values_to", {
   )
 })
 
-test_that("vector column can parse scalar list", {
+test_that("tib_vector can parse scalar list", {
   spec <- tib_int_vec("x", input_form = "scalar_list")
   expect_equal(
     tib(list(x = list(1, NULL, 3)), spec),
@@ -328,24 +350,29 @@ test_that("vector column can parse scalar list", {
     tibble(x = list_of(1:2, integer()))
   )
 
-  tspec_object <- spec
-  tspec_object$input_form <- "object"
+  # checks that input is a list
   expect_snapshot({
     (expect_error(tib(list(x = 1), spec)))
-    (expect_error(tib(list(x = 1), tspec_object)))
   })
 
+  # each element must have size 1
   expect_snapshot({
     (expect_error(tib(list(x = list(1, 1:2)), spec)))
     (expect_error(tib(list(x = list(integer())), spec)))
   })
 
+  # each element must have size 1 also after encountering a NULL
+  expect_snapshot({
+    (expect_error(tib(list(x = list(NULL, 1, 1:2)), spec)))
+  })
+
+  # each element must have the correct type
   expect_snapshot({
     (expect_error(tib(list(x = list(1, "a")), spec)))
   })
 })
 
-test_that("vector column can parse object", {
+test_that("tib_vector can parse object", {
   spec <- tib_int_vec("x", input_form = "object")
   expect_equal(
     tib(list(x = list(a = 1, b = NULL, c = 3)), spec),
@@ -366,9 +393,34 @@ test_that("vector column can parse object", {
     tib(list(x = list(a = 1, a = 2)), spec),
     tibble(x = list_of(c(1L, 2L)))
   )
+
+  # must be a list
+  expect_snapshot({
+    (expect_error(tib(list(x = 1), spec)))
+  })
+
+  # each element must have size 1
+  expect_snapshot({
+    (expect_error(tib(list(x = list(a = 1, b = 1:2)), spec)))
+  })
 })
 
-test_that("vector column creates tibble with names_to", {
+test_that("tib_vector creates tibble with names_to", {
+  # TODO not quite sure whether `val` should be named
+  expect_equal(
+    tib(list(x = c(a = 1L, b = 2L)), tib_int_vec("x", names_to = "name", values_to = "val")),
+    tibble(x = list_of(
+      tibble(name = c("a", "b"), val = c(a = 1L, b = 2L)),
+      .ptype = tibble(name = character(), val = integer())
+    ))
+  )
+
+  # can handle missing names
+  expect_equal(
+    tib(list(x = 1:2), tib_int_vec("x", names_to = "name", values_to = "val")),
+    tibble(x = list_of(tibble(name = NA_character_, val = 1:2)))
+  )
+
   spec <- tib_int_vec("x", input_form = "object", values_to = "val", names_to = "name")
   expect_equal(
     tib(list(x = list(a = 1, b = NULL)), spec),
@@ -396,7 +448,7 @@ test_that("vector column creates tibble with names_to", {
   )
 })
 
-test_that("list column works", {
+test_that("tib_variant works", {
   # can parse
   expect_equal(
     tibblify(
@@ -453,13 +505,22 @@ test_that("list column works", {
   expect_equal(
     tibblify(
       list(list(x = c(TRUE, FALSE)), list(x = 1)),
-      tspec_df(x = tib_variant("x", required = FALSE, transform = ~ length(.x)))
+      tspec_df(x = tib_variant("x", required = FALSE, transform = lengths))
+    ),
+    tibble(x = c(2L, 1L))
+  )
+
+  # elt_transform works
+  expect_equal(
+    tibblify(
+      list(list(x = c(TRUE, FALSE)), list(x = 1)),
+      tspec_df(x = tib_variant("x", required = FALSE, elt_transform = ~ length(.x)))
     ),
     tibble(x = list(2, 1))
   )
 })
 
-test_that("df column works", {
+test_that("tib_row works", {
   # can parse
   expect_equal(
     tibblify(
@@ -487,17 +548,29 @@ test_that("df column works", {
   expect_equal(
     tibblify(
       list(
-        list(x = list(a = TRUE)),
+        list(x = list(a = 1)),
         list(x = list()),
         list()
       ),
-      tspec_df(x = tib_row("x", .required = FALSE, a = tib_lgl("a", required = FALSE)))
+      tspec_df(x = tib_row("x", .required = FALSE, a = tib_int("a", required = FALSE, fill = -1)))
     ),
-    tibble(x = tibble(a = c(TRUE, NA, NA)))
+    tibble(x = tibble(a = c(1L, -1L, -1L)))
+  )
+
+  # can handle NULL
+  # TODO not quite sure about this
+  expect_equal(
+    tibblify(
+      list(
+        list(x = NULL)
+      ),
+      tspec_df(x = tib_row("x", .required = FALSE, a = tib_int("a", required = FALSE, fill = -1)))
+    ),
+    tibble(x = tibble(a = -1L))
   )
 })
 
-test_that("list of df column works", {
+test_that("tib_df works", {
   # can parse
   expect_equal(
     tibblify(
@@ -520,7 +593,7 @@ test_that("list of df column works", {
           list(a = TRUE),
           list(a = FALSE)
         )),
-        list()
+        list(a = 1)
       ),
       tspec_df(x = tib_df("x", a = tib_lgl("a")))
     )))
@@ -563,9 +636,45 @@ test_that("list of df column works", {
     ),
     tibble(x = list_of(NULL, .ptype = tibble(a = logical())))
   )
+
+  # can handle every `tib_*()`
+  dtt <- vctrs::new_datetime(1)
+  x <- list(
+    atomic_scalar = 1L,
+    scalar = dtt,
+    vector = 1:2,
+    variant = list(1L, "a"),
+    row = list(a = 1, b = "b"),
+    df = list(list(x = 1))
+  )
+  spec <- tspec_object(
+    tib_df(
+      "df",
+      tib_int("atomic_scalar"),
+      tib_scalar("scalar", dtt),
+      tib_int_vec("vector"),
+      tib_variant("variant"),
+      tib_row("row", tib_int("a"), tib_chr("b")),
+      tib_df("df", tib_int("x")),
+    )
+  )
+
+  expect_equal(
+    tibblify(list(df = list(x)), spec),
+    list(
+      df = tibble(
+        atomic_scalar = 1L,
+        scalar = dtt,
+        vector = list_of(1:2),
+        variant = list(list(1L, "a")),
+        row = tibble(a = 1L, b = "b"),
+        df = list_of(tibble(x = 1L))
+      )
+    )
+  )
 })
 
-test_that("list of df column can use names_to", {
+test_that("tib_df can use names_to", {
   spec <- tspec_df(x = tib_df("x", a = tib_lgl("a"), .names_to = "name"))
   expect_equal(
     tibblify(
@@ -674,6 +783,30 @@ test_that("nested keys work", {
     ),
     tibble(xyz = 1)
   )
+
+  spec <- tspec_df(
+    afk = tib_chr(c("a", "f", "k")),
+    ael = tib_int(c("a", "e", "l")),
+    aek = tib_dbl(c("a", "e", "k"))
+  )
+
+  out <- tibblify(
+    list(
+      list(a = list(
+        e = list(k = -0.1, l = 1L),
+        f = list(k = "a")
+      )),
+      list(a = list(
+        e = list(k = +0.2, l = 2L),
+        f = list(k = "b")
+      ))
+    ),
+    spec
+  )
+  expect_equal(
+    out,
+    tibble(afk = c("a", "b"), ael = 1:2, aek = c(-0.1, +0.2))
+  )
 })
 
 test_that("empty spec works", {
@@ -903,6 +1036,21 @@ test_that("tspec_object() works", {
       )
     )
   )
+
+  # can handle sub collector + different order
+  spec3 <- tspec_object(
+    tib_chr("c"),
+    bb = tib_chr(c("b", "b")),
+    tib_chr("a"),
+    ba = tib_chr(c("b", "a")),
+  )
+  expect_equal(
+    tibblify(
+      list(a = "a", c = "c", b = list(a = "ba", b = "bb")),
+      spec3
+    ),
+    list(c = "c", bb = "bb", a = "a", ba = "ba")
+  )
 })
 
 test_that("spec_replace_unspecified works", {
@@ -997,16 +1145,13 @@ test_that("colmajor: names are checked", {
   })
 })
 
-test_that("colmajor: scalar column works", {
+test_that("colmajor: tib_scalar works", {
   dtt <- vctrs::new_datetime(1)
 
   # can parse
   expect_equal(tib_cm(x = TRUE, tib_lgl("x")), tibble(x = TRUE))
   expect_equal(tib_cm(x = dtt, tib_scalar("x", dtt)), tibble(x = dtt))
 
-  # errors if required but absent
-  # expect_snapshot((expect_error(tib2(x = 1:3, tib_lgl("y")))))
-
   # errors if bad type
   expect_snapshot({
     (expect_error(tib_cm(x = "a", tib_lgl("x"))))
@@ -1022,34 +1167,9 @@ test_that("colmajor: scalar column works", {
     tib_cm(x = dtt, tib_scalar("x", dtt, transform = ~ .x + 1)),
     tibble(x = vctrs::new_datetime(2))
   )
-
-  skip("Unclear if required and default makes sense for colmajor")
-  # fallback default works
-  expect_equal(
-    tib2(x = 1:2, tib_int("x"), tib_lgl("y", required = FALSE)),
-    tibble(x = 1:2, y = NA)
-  )
-  expect_equal(
-    tib2(x = 1:2, tib_int("x"), tib_scalar("y", dtt, required = FALSE)),
-    tibble(x = 1:2, y = vctrs::new_datetime(NA_real_))
-  )
-
-  # use NA if NULL
-  expect_equal(tib(list(x = NULL), tib_lgl("x", required = FALSE, fill = FALSE)), tibble(x = NA))
-  expect_equal(
-    tib(list(x = NULL), tib_scalar("x", vec_ptype(dtt), required = FALSE, fill = dtt)),
-    tibble(x = vec_init(dtt))
-  )
-
-  # specified default works
-  expect_equal(tib(list(), tib_lgl("x", required = FALSE, fill = FALSE)), tibble(x = FALSE))
-  expect_equal(
-    tib(list(), tib_scalar("x", vec_ptype(dtt), required = FALSE, fill = dtt)),
-    tibble(x = dtt)
-  )
 })
 
-test_that("colmajor: record objects work", {
+test_that("colmajor: tib_scalar with record objects work", {
   x_rcrd <- as.POSIXlt(Sys.time(), tz = "UTC")
 
   expect_equal(
@@ -1065,7 +1185,7 @@ test_that("colmajor: record objects work", {
   )
 })
 
-test_that("colmajor: scalar columns respect ptype_inner", {
+test_that("colmajor: tib_scalar respect ptype_inner", {
   f <- function(x) {
     stopifnot(is.character(x))
     as.Date(x)
@@ -1090,17 +1210,13 @@ test_that("colmajor: scalar columns respect ptype_inner", {
   )
 })
 
-test_that("colmajor: vector column works", {
+test_that("colmajor: tib_vector works", {
   dtt <- vctrs::new_datetime(1)
 
   # can parse
   expect_equal(tib_cm(tib_lgl_vec("x"), x = list(c(TRUE, FALSE))), tibble(x = list_of(c(TRUE, FALSE))))
   expect_equal(tib_cm(tib_vector("x", dtt), x = list(c(dtt, dtt + 1))), tibble(x = list_of(c(dtt, dtt + 1))))
 
-  # errors if required but absent
-  # expect_snapshot((expect_error(tib_cm(tib_lgl_vec("x"), list()))))
-  # expect_snapshot((expect_error(tib_cm(tib_vector("x", dtt), list()))))
-
   # errors if bad type
   expect_snapshot({
     # not a list
@@ -1111,25 +1227,29 @@ test_that("colmajor: vector column works", {
 
   # transform works
   expect_equal(
-    tib_cm(tib_lgl_vec("x", transform = ~ !.x), x = list(c(TRUE, FALSE))),
+    tib_cm(
+      tib_lgl_vec("x", transform = ~ vctrs::new_list_of(purrr::map(.x, `!`), logical())),
+      x = list(c(TRUE, FALSE))
+    ),
     tibble(x = list_of(c(FALSE, TRUE)))
   )
   expect_equal(
-    tib_cm(tib_vector("x", dtt, transform = ~ .x + 1), x = list(c(dtt - 1, dtt))),
-    tibble(x = list_of(c(dtt, dtt + 1)))
+    tib_cm(
+      tib_vector("x", dtt, transform = ~ vctrs::new_list_of(purrr::map(.x, ~ .x + 1), vctrs::vec_ptype(dtt))),
+      x = list(c(dtt - 1, dtt))
+    ),
+    tibble(x = list_of(c(dtt, dtt + 1))),
+    ignore_attr = c("tzone", "waldo_opts", "tib_spec")
   )
 
-  skip("Unclear if required and default makes sense for colmajor")
-  # fallback default works
-  expect_equal(tib(list(), tib_lgl_vec("x", required = FALSE)), tibble(x = list_of(NULL, .ptype = logical())))
-  expect_equal(tib(list(), tib_vector("x", dtt, required = FALSE)), tibble(x = list_of(NULL, .ptype = vctrs::new_datetime())))
-
-  # specified default works
-  expect_equal(tib(list(), tib_lgl_vec("x", required = FALSE, fill = c(TRUE, FALSE))), tibble(x = list_of(c(TRUE, FALSE))))
-  expect_equal(tib(list(), tib_vector("x", dtt, required = FALSE, fill = c(dtt, dtt + 1))), tibble(x = list_of(c(dtt, dtt + 1))))
+  # elt_transform works
+  expect_equal(
+    tib_cm(tib_lgl_vec("x", elt_transform = ~ !.x), x = list(c(TRUE, FALSE))),
+    tibble(x = list_of(c(FALSE, TRUE)))
+  )
 })
 
-test_that("list column works", {
+test_that("colmajor: tib_variant works", {
   # can parse
   expect_equal(
     tib_cm(tib_variant("x"), x = list(TRUE, 1)),
@@ -1139,14 +1259,23 @@ test_that("list column works", {
   # transform works
   expect_equal(
     tib_cm(
-      tib_variant("x", required = FALSE, transform = ~ length(.x)),
+      tib_variant("x", required = FALSE, transform = lengths),
+      x = list(c(TRUE, FALSE), 1)
+    ),
+    tibble(x = c(2L, 1L))
+  )
+
+  # elt_transform works
+  expect_equal(
+    tib_cm(
+      tib_variant("x", required = FALSE, elt_transform = ~ length(.x)),
       x = list(c(TRUE, FALSE), 1)
     ),
     tibble(x = list(2, 1))
   )
 })
 
-test_that("row works", {
+test_that("colmajor: tib_row works", {
   # can parse
   expect_equal(
     tib_cm(
@@ -1164,34 +1293,9 @@ test_that("row works", {
     ),
     tibble(x = tibble(row = tibble(int = 1:2, chr_vec = list_of("a", c("b", "c")))))
   )
-
-  skip("Unclear if required and default makes sense for colmajor")
-  # errors if required but absent
-  expect_snapshot(
-    (expect_error(tibblify(
-      list(
-        list(x = list(a = TRUE)),
-        list()
-      ),
-      tspec_df(x = tib_row("x", a = tib_lgl("a")))
-    )))
-  )
-
-  # fallback default works
-  expect_equal(
-    tibblify(
-      list(
-        list(x = list(a = TRUE)),
-        list(x = list()),
-        list()
-      ),
-      tspec_df(x = tib_row("x", .required = FALSE, a = tib_lgl("a", required = FALSE)))
-    ),
-    tibble(x = tibble(a = c(TRUE, NA, NA)))
-  )
 })
 
-test_that("list of df column works", {
+test_that("colmajor: tib_df works", {
   # can parse
   expect_equal(
     tib_cm(
@@ -1224,39 +1328,19 @@ test_that("list of df column works", {
     ))
   )
 
-  skip("Unclear if required and default makes sense for colmajor")
-  # errors if required but absent
-  expect_snapshot(
-    (expect_error(tibblify(
-      list(
-        list(x = list(
-          list(a = TRUE),
-          list(a = FALSE)
-        )),
-        list()
-      ),
-      tspec_df(x = tib_df("x", a = tib_lgl("a")))
+  # Field x$x[[2]]$b has size 1
+  expect_snapshot({
+    (expect_error(tib_cm(
+      tib_df("x", tib_int("a"), tib_chr("b")),
+      x = list(
+        list(b = c("a", "b"), a = 1:2),
+        list(a = character(), b = "c")
+      )
     )))
-  )
-
-  # fallback default works
-  expect_equal(
-    tibblify(
-      list(
-        list(x = list(
-          list(a = TRUE),
-          list(a = FALSE)
-        )),
-        list(x = list()),
-        list()
-      ),
-      tspec_df(x = tib_df("x", .required = FALSE, a = tib_lgl("a", required = FALSE)))
-    ),
-    tibble(x = list_of(tibble(a = c(TRUE, FALSE)), tibble(a = logical()), NULL))
-  )
+  })
 })
 
-test_that("tibble with list columns work - #43", {
+test_that("colmajor: tibble with list columns work - #43", {
   x <- tibble::tibble(x = list(1:3, NULL, 1:2))
   expect_equal(
     tibblify(x, tspec_df(x = tib_int_vec("x"), .input_form = "colmajor")),
@@ -1271,7 +1355,7 @@ test_that("tibble with list columns work - #43", {
   )
 })
 
-test_that("nested keys work", {
+test_that("colmajor: nested keys work", {
   spec <- tspec_df(
     xyz = tib_int(c("x", "y", "z")),
     .input_form = "colmajor"
@@ -1280,36 +1364,9 @@ test_that("nested keys work", {
     tibblify(list(x = list(y = list(z = 1))), spec),
     tibble(xyz = 1)
   )
-
-  skip("Unclear if required and default makes sense for colmajor")
-  spec2 <- spec
-  spec2$fields$xya <-tib_int(c("x", "y", "a"), required = FALSE, fill = 2)
-  expect_equal(
-    tibblify(list(x = list(y = list(z = 1))), spec2),
-    tibble(xyz = 1, xya = NA_integer_)
-  )
 })
 
-test_that("default for same key collector works", {
-  skip("undecided whether this should work")
-  expect_equal(
-    tibblify(
-      list(
-        x = list(),
-        y = 1:2
-      ),
-      tspec_df(
-        tib_int("y"),
-        a = tib_int(c("x", "a"), required = FALSE),
-        b = tib_int(c("x", "b"), required = FALSE),
-        .input_form = "colmajor"
-      )
-    ),
-    tibble(a = c(1L, NA), b = c(2L, NA))
-  )
-})
-
-test_that("empty spec works", {
+test_that("colmajor: empty spec works", {
   expect_equal(
     tibblify(
       set_names(list()),
@@ -1327,64 +1384,254 @@ test_that("empty spec works", {
   )
 })
 
-test_that("errors if n_rows cannot be calculated", {
+test_that("colmajor: errors if field is absent", {
   expect_snapshot({
-    # after key in alphabet
     (expect_error(tib_cm(tib_int("y"), x = list(b = 1:3))))
-    # before key in alphabet
-    (expect_error(tib_cm(tib_int("a"), x = list(b = 1:3))))
+    # errors even if `required = FALSE`
+    (expect_error(tib_cm(tib_int("x"), tib_int("y", required = FALSE), x = list(b = 1:3))))
   })
 })
 
-test_that("colmajor can calculate size", {
+test_that("colmajor: can calculate size", {
+  # atomic scalar
+  expect_equal(
+    tibblify(
+      list(x = 1:2),
+      tspec_df(tib_int("x"), .input_form = "colmajor")
+    ),
+    tibble(x = 1:2)
+  )
+
+  # record objects
+  x_rcrd <- as.POSIXlt(Sys.time(), tz = "UTC")
+  expect_equal(
+    tibblify(
+      list(x = c(x_rcrd, x_rcrd + 1)),
+      tspec_df(tib_scalar("x", x_rcrd), .input_form = "colmajor")
+    ),
+    tibble(x = c(x_rcrd, x_rcrd + 1))
+  )
+
+  # vector
+  expect_equal(
+    tibblify(
+      list(x = list(1:4, 1:3)),
+      tspec_df(tib_int_vec("x"), .input_form = "colmajor")
+    ),
+    tibble(x = list_of(1:4, 1:3))
+  )
+
+  # row
+  expect_equal(
+    tibblify(
+      list(x = list(a = 1:2)),
+      tspec_df(tib_row("x", tib_int("a")), .input_form = "colmajor")
+    ),
+    tibble(x = tibble(a = 1:2))
+  )
+
+  # df
   expect_equal(
     tibblify(
       list(
-        x = NULL,
-        y = 1:2
+        x = list(
+          list(a = 1:2),
+          list(a = 1)
+        )
       ),
-      tspec_df(tib_int("x"), tib_int("y"), .input_form = "colmajor")
+      tspec_df(tib_df("x", tib_int("a")), .input_form = "colmajor")
     ),
-    tibble(x = NA_integer_, y = 1:2)
+    tibble(x = list_of(tibble(a = 1:2), tibble(a = 1)))
   )
 
   expect_snapshot(
-    expect_error(
+    (expect_error(
       tibblify(
         list(row = "a"),
         tspec_df(tib_row("row", tib_int("x")), .input_form = "colmajor")
       )
-    )
-  )
-
-  skip("undecided whether this should work")
-  expect_equal(
-    tibblify(
-      list(
-        row = NULL,
-        x = 1:2
-      ),
-      tspec_df(tib_row("row", tib_int("x")), tib_int("x"), .input_form = "colmajor")
-    ),
-    tibble(x = NA_integer_, y = 1:2)
+    ))
   )
 })
 
-test_that("colmajor checks size", {
-  tib_cm <- function(col1, col2, x, y) {
-    tibblify(
-      list(x = x, y = y),
-      tspec_df(col1, col2, .input_form = "colmajor")
-    )
-  }
-
-  spec_cm <- function(...) {
-    tspec_df(.input_form = "colmajor", ...)
-  }
+test_that("colmajor: errors on NULL value", {
+  expect_snapshot({
+    (expect_error(tib_cm(tib_int("x"), x = NULL)))
+  })
+})
 
+test_that("colmajor: checks size", {
   expect_snapshot({
     (expect_error(tib_cm(tib_int("x"), tib_int("y"), x = 1:2, y = 1:3)))
     (expect_error(tib_cm(tib_int("x"), tib_row("y", tib_int("x")), x = 1:2, y = list(x = 1:3))))
     (expect_error(tib_cm(tib_int("x"), tib_int_vec("y"), x = 1:2, y = list(1))))
   })
 })
+
+
+# recursive ---------------------------------------------------------------
+
+test_that("recursive: works", {
+  spec <- tspec_recursive(
+    tib_int("id"),
+    tib_chr("name"),
+    .children = "children"
+  )
+  spec2 <- tspec_object(
+    tib_recursive(
+      "data",
+      tib_int("id"),
+      tib_chr("name"),
+      .children = "children"
+    )
+  )
+
+  expected <- tibble(id = 1L, name = "a", children = list(NULL))
+  expect_equal(
+    tibblify(list(list(id = 1, name = "a")), spec),
+    expected
+  )
+
+  expect_equal(
+    tibblify(list(data = list(list(id = 1, name = "a"))), spec2),
+    list(data = expected)
+  )
+
+  expect_equal(
+    tibblify(list(list(id = 1, name = "a", children = NULL)), spec),
+    tibble(id = 1L, name = "a", children = list(NULL))
+  )
+
+  # TODO not quite sure
+  expect_equal(
+    tibblify(list(list(id = 1, name = "a", children = list())), spec),
+    tibble(
+      id = 1L,
+      name = "a",
+      children = list(tibble(id = integer(), name = character(), children = list()))
+    )
+  )
+
+  # deeply nested works
+  x <- list(
+    list(id = 1, name = "a", children = list(
+      list(id = 11, name = "aa"),
+      list(id = 12, name = "ab", children = list(
+        list(id = 121, name = "aba")
+      ))
+    )),
+    list(id = 2, name = "b", children = list(
+      list(id = 21, name = "ba", children = list(
+        list(id = 121, name = "bba", children = list(
+          list(id = 1211, name = "bbaa")
+        ))
+      )),
+      list(id = 22, name = "bb")
+    ))
+  )
+
+  expected <- tibble(
+    id = 1:2,
+    name = c("a", "b"),
+    children = list(
+      tibble(
+        id = 11:12,
+        name = c("aa", "ab"),
+        children = list(
+          NULL,
+          tibble(id = 121, name = "aba", children = list(NULL))
+        )
+      ),
+      tibble(
+        id = 21:22,
+        name = c("ba", "bb"),
+        children = list(
+          tibble(
+            id = 121,
+            name = "bba",
+            children = list(
+              tibble(id = 1211, name = "bbaa", children = list(NULL))
+            )
+          ),
+          NULL
+        )
+      )
+    )
+  )
+
+  expect_equal(tibblify(x, spec), expected)
+  expect_equal(tibblify(list(data = x), spec2), list(data = expected))
+
+  # tibble input
+  expect_equal(tibblify(expected, spec), expected)
+  # colmajor
+  spec_colmajor <- tspec_recursive(
+    tib_int("id"),
+    tib_chr("name"),
+    .children = "children",
+    .input_form = "colmajor"
+  )
+
+  x_cm <- list(
+    id = 1:2,
+    name = c("a", "b"),
+    children = list(
+      list(
+        id = 11:12,
+        name = c("aa", "ab"),
+        children = list(
+          NULL,
+          list(id = 121, name = "aba", children = list(NULL))
+        )
+      ),
+      list(
+        id = 21:22,
+        name = c("ba", "bb"),
+        children = list(
+          list(
+            id = 121,
+            name = "bba",
+            children = list(
+              list(id = 1211, name = "bbaa", children = list(NULL))
+            )
+          ),
+          NULL
+        )
+      )
+    )
+  )
+  expect_equal(tibblify(x_cm, spec_colmajor), expected)
+
+  x2 <- x
+  x2[[1]]$children[[2]]$children[[1]]$id <- "does not work"
+  expect_snapshot(
+    (expect_error(tibblify(x2, spec)))
+  )
+})
+
+test_that("recursive: empty input works", {
+  spec <- tspec_recursive(
+    tib_int("id"),
+    tib_chr("name"),
+    .children = "children"
+  )
+
+  expect_equal(
+    tibblify(list(), spec),
+    tibble(id = integer(), name = character(), children = list())
+  )
+})
+
+test_that("recursive: empty spec works", {
+  expect_equal(
+    tibblify(list(), tspec_recursive(.children = "children")),
+    tibble(children = list())
+  )
+})
+
+test_that("recursive: .children_to works", {
+  expect_equal(
+    tibblify(list(), tspec_recursive(.children = "children", .children_to = "child_col")),
+    tibble(child_col = list())
+  )
+})
